package org.ow2.orchestra.env.xml;

import java.io.InputStream;
import java.net.URL;

import org.ow2.orchestra.env.Context;
import org.ow2.orchestra.env.EnvironmentFactory;
import org.ow2.orchestra.env.PvmEnvironmentFactory;
import org.ow2.orchestra.env.WireContext;
import org.ow2.orchestra.env.WireDefinition;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.util.stream.StreamSource;
import org.ow2.orchestra.xml.Parse;
import org.ow2.orchestra.xml.Parser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

public class EnvironmentFactoryParser extends Parser {

  private static final String ENVIRONMENT_ORCHESTRA_XSD = "environment.xsd";
  public static final String CLASSPATH_URL = "classpath://";

  private WireParser environmentFactoryXmlParser = new WireParser();
  private WireParser environmentXmlParser = new WireParser();

  private static final EnvironmentFactoryParser INSTANCE = new EnvironmentFactoryParser();

  public EnvironmentFactoryParser() {
    this.classLoader = Thread.currentThread().getContextClassLoader();
  }

  public static EnvironmentFactoryParser getInstance() {
    return EnvironmentFactoryParser.INSTANCE;
  }

  /**
   * parses the given stream source with {@link #getInstance() the current
   * parser}.
   */
  static EnvironmentFactory parse(final StreamSource streamSource) {
    final EnvironmentFactory environmentFactory = (EnvironmentFactory) EnvironmentFactoryParser.getInstance()
        .createParse().setStreamSource(streamSource).execute().checkProblems(
            "environment configuration " + streamSource).getDocumentObject();
    return environmentFactory;
  }

  @Override
  public Object parseDocument(final Document document, final Parse parse) {
    final Element documentElement = document.getDocumentElement();

    // if the default environment factory was already set in the parse
    PvmEnvironmentFactory pvmEnvironmentFactory = (PvmEnvironmentFactory) parse
        .getDocumentObject();
    if (pvmEnvironmentFactory == null) {
      pvmEnvironmentFactory = new PvmEnvironmentFactory();
      parse.setDocumentObject(pvmEnvironmentFactory);
    }

    final WireDefinition environmentFactoryWireDefinition = this.getApplicationWireDefinition(
        documentElement, parse);
    final WireDefinition environmentWireDefinition = this.getBlockWireDefinition(
        documentElement, parse);

    // create the application wire context from the definition
    final WireContext environmentFactoryWireContext = new WireContext(
        environmentFactoryWireDefinition,
        Context.CONTEXTNAME_ENVIRONMENT_FACTORY);
    // propagate the parser classloader to the application context
    environmentFactoryWireContext.setClassLoader(this.classLoader);

    // configure the default environment factory
    pvmEnvironmentFactory
        .setEnvironmentFactoryCtxWireContext(environmentFactoryWireContext);
    pvmEnvironmentFactory
        .setEnvironmentCtxWireDefinition(environmentWireDefinition);

    parse.setDocumentObject(pvmEnvironmentFactory);

    return pvmEnvironmentFactory;
  }

  WireDefinition getApplicationWireDefinition(final Element documentElement,
      final Parse parse) {
    final Element applicationElement = XmlUtil.element(documentElement,
        Context.CONTEXTNAME_ENVIRONMENT_FACTORY);
    if (applicationElement != null) {
      return (WireDefinition) this.environmentFactoryXmlParser.parseDocumentElement(
          applicationElement, parse);
    }
    return null;
  }

  WireDefinition getBlockWireDefinition(final Element documentElement, final Parse parse) {
    final Element blockElement = XmlUtil.element(documentElement,
        Context.CONTEXTNAME_ENVIRONMENT);
    if (blockElement != null) {
      return (WireDefinition) this.environmentXmlParser.parseDocumentElement(
          blockElement, parse);
    }
    return null;
  }

  public WireParser getEnvironmentFactoryXmlParser() {
    return this.environmentFactoryXmlParser;
  }

  public void setEnvironmentFactoryXmlParser(final WireParser applicationWireXmlParser) {
    this.environmentFactoryXmlParser = applicationWireXmlParser;
  }

  public WireParser getEnvironmentXmlParser() {
    return this.environmentXmlParser;
  }

  public void setEnvironmentXmlParser(final WireParser blockWireXmlParser) {
    this.environmentXmlParser = blockWireXmlParser;
  }

  public static EnvironmentFactory parseEnvironmentFactoryFromXmlString(final String xmlString) {
    final Parse parse = EnvironmentFactoryParser.getInstance().createParse();
    final EnvironmentFactory factory = (EnvironmentFactory) parse.setString(xmlString).execute().getDocumentObject();
    parse.checkProblems("environment");
    return factory;
  }

  public static synchronized EnvironmentFactory parseEnvironmentFactoryFromURL(final URL url) {
    if (url == null) {
      throw new OrchestraRuntimeException("No environment xml found");
    }
    final Parse parse = EnvironmentFactoryParser.getInstance().createParse();
    final PvmEnvironmentFactory factory = (PvmEnvironmentFactory) parse.setUrl(url).execute().getDocumentObject();
    parse.checkProblems("environment");
    return factory;
  }

  public static synchronized EnvironmentFactory parseEnvironmentFactoryFromResource(final String resource) {
    Misc.checkArgsNotNull(resource);
    final Parse parse = EnvironmentFactoryParser.getInstance().createParse();
    final EnvironmentFactory factory = (EnvironmentFactory) parse.setResource(resource).execute().getDocumentObject();
    parse.checkProblems("environment");
    return factory;
  }

  /* (non-Javadoc)
   * @see org.ow2.orchestra.pvm.internal.xml.Parser#getSchemaSources()
   */
  @Override
  protected String[] getSchemaSources() {
    return new String[] {EnvironmentFactoryParser.CLASSPATH_URL + EnvironmentFactoryParser.ENVIRONMENT_ORCHESTRA_XSD};
  }

  /*
   * (non-Javadoc)
   *
   * @see org.ow2.orchestra.xml.Parser#resolveEntity(java.lang.String,
   * java.lang.String)
   */
  @Override
  public InputSource resolveEntity(final String publicId, final String systemId) {
    if (systemId.startsWith(EnvironmentFactoryParser.CLASSPATH_URL)) {
      final String localpart = systemId.substring(EnvironmentFactoryParser.CLASSPATH_URL.length());
      final InputStream inStream = Thread.currentThread().getContextClassLoader().getResourceAsStream(localpart);
      if (inStream != null) {
        return new InputSource(inStream);
      }
    }
    return super.resolveEntity(publicId, systemId);
  }
}
