/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.lang.function;

import java.io.ByteArrayInputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;

import org.jaxen.Context;
import org.jaxen.ContextSupport;
import org.jaxen.Function;
import org.jaxen.FunctionCallException;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.lang.jaxen.BpelVariableContext;
import org.ow2.orchestra.util.ReflectUtil;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * The <strong>doXslTranformationFunction</strong> function.
 *
 * @author Charles Souillard
 * @author Guillaume Porcher
 *
 * @version $Revision: 1.5 $ $Date: 2007/07/26 00:16:08 $
 */
public class DoXslTransformationFunction implements Function {

  /**
   * Extracts the value of a global property from a variable.
   *
   * @param context
   *            the context at the point in the expression when the function is
   *            called
   * @param args
   *            a list containing the variable name and the property qualified
   *            name
   * @return the property vlaue
   * @throws FunctionCallException
   *             if <code>args</code> does not have exactly two items
   */
  public Object call(final Context context, final List args) throws FunctionCallException {
    if (args.size() < 2) {
      throw new FunctionCallException(
          "Wrong number of argument for the bpel:doXslTransform function.");
    }
    final ContextSupport sup = context.getContextSupport();
    final BpelVariableContext variableContext = (BpelVariableContext) sup
        .getVariableContext();
    final Map<String, byte[]> styleSheets = variableContext.getStyleSheets();
    final byte[] styleSheetBytes = styleSheets.get(args.get(0));
    if (styleSheetBytes == null) {
      throw BpelFaultException.generateBpelFaultException(BpelFault.xsltStylesheetNotFound);
    }
    Node xpathNodeSet;
    try {
      xpathNodeSet = (Node) args.get(1);
    } catch (final ClassCastException e) {
      try {
        final List<Node> nodeList = (List) args.get(1);
        xpathNodeSet = nodeList.get(0);
      } catch (final Exception e1) {
        throw new FunctionCallException(
          "The second parameter of bpel:doXslTransform must be a valid Node.");
      }
    }
    StreamSource styleSheet = null;
    try {
      styleSheet = new StreamSource(new ByteArrayInputStream(styleSheetBytes));
      final TransformerFactory tFactory = XmlUtil.getTransformerFactory();
      final Transformer transformer = tFactory.newTransformer(styleSheet);

      // PARAMETERS
      int i = 2;
      while (i < args.size()) {
        Object param = args.get(i + 1);
        if (param instanceof Collection< ? >) {
          final Collection< ? > parList = (Collection< ? >) param;
          if (parList.size() == 1) {
            param = parList.iterator().next();
          }
        }
        // convert Node to DTM for sun xsltc
        param = this.useDTMManagerOnParam(transformer, param,
            "com.sun.org.apache.xalan.internal.xsltc.trax.TransformerImpl",
            "com.sun.org.apache.xalan.internal.xsltc.dom.XSLTCDTMManager",
            "com.sun.org.apache.xml.internal.dtm.DTMWSFilter");
        // convert Node to DTM for xalan xsltc
        param = this.useDTMManagerOnParam(transformer, param,
            "org.apache.xalan.xsltc.trax.TransformerImpl",
            "org.apache.xalan.xsltc.dom.XSLTCDTMManager",
            "org.apache.xml.dtm.DTMWSFilter");


        transformer.setParameter((String) args.get(i), param);
        i += 2;
      }

      // SOURCE
      final Document document = XmlUtil.getNewDocument();
      document.appendChild(document.importNode(xpathNodeSet, true));
      final DOMSource source = new DOMSource(document);

      // RESULT
      final DOMResult result = new DOMResult();
      // TRANSFORMATION
      transformer.transform(source, result);
      final Document doc = (Document) result.getNode();
      final List<Object> listReturned = new ArrayList<Object>();
      listReturned.add(doc.getFirstChild());

      return listReturned;

    } catch (final Exception e) {
      throw new OrchestraRuntimeException(
          "Exception caught when doing xsl transformation : ", e);
    } finally {
      try {
        styleSheet.getInputStream().close();
      } catch (final Exception e) {
        // nothing
      }
    }
  }

  /**
   * Convert param object to DTM parameter for xsltc implementations.
   * Using directly the DOM element param does not work for these implementations (it's converted to string).
   * @param transformer
   * @param param
   * @param transformerClassName
   * @param dtmManagerClassName
   * @param dtmWSFilterClassName
   * @return
   */
  private Object useDTMManagerOnParam(
      final Transformer transformer,
      final Object param,
      final String transformerClassName,
      final String dtmManagerClassName,
      final String dtmWSFilterClassName) {
    try {
      final Class< ? > sunTransformerClass = Class.forName(transformerClassName, true, transformer.getClass().getClassLoader());
      if (sunTransformerClass.isInstance(transformer)) {
        if (param instanceof Node) {
          final Class< ? > xsltCDTMManagerClass =
            Class.forName(dtmManagerClassName, true, transformer.getClass().getClassLoader());
          final Object instance = xsltCDTMManagerClass.newInstance();

          final Class< ? > dtmWSFilterClass =
            Class.forName(dtmWSFilterClassName, true, transformer.getClass().getClassLoader());
          final Method m = ReflectUtil.getMethod(
              xsltCDTMManagerClass,
              "getDTM",
              new Class< ? >[]{Source.class, Boolean.TYPE, dtmWSFilterClass, Boolean.TYPE, Boolean.TYPE});
          return ReflectUtil.invoke(m, instance, new Object[]{new DOMSource((Node)param), false, null, false, false});
        }
      }
    } catch (final Exception e) {
        // Not the transformer instance, ignore.
    }
    return param;
  }
}
