/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.osgi;

import java.io.File;
import java.io.FileInputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.felix.ipojo.ComponentInstance;
import org.apache.felix.ipojo.Factory;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;
import org.ow2.orchestra.util.Misc;

/**
 * @author Guillaume Porcher
 * 
 */
public class OrchestraWSEngineActivator extends ExtensionActivator {

  /**
   * @author Guillaume Porcher
   * 
   */
  private static final class EnvironmentProvider implements EnvironmentProviderService {

    private final URL envUrl;
    private final Bundle envBundle;

    EnvironmentProvider(final URL envUrl, final BundleContext bundleContext) {
      this.envUrl = envUrl;
      this.envBundle = bundleContext.getBundle();
    }
    public URL getEnvironmentURL() {
      return this.envUrl;
    }

    public Bundle getEnvironmentBundle() {
      return this.envBundle;
    }
  }

  private static final Logger LOG = Logger.getLogger(OrchestraWSEngineActivator.class.getName());

  // Default file to list user configurations
  private static final String DEFAULT_USER_CONFIG_FILE = "conf/user-environments.properties";
  // Default environment for deployed services
  private static final String DEFAULT_ENV_CONFIG_FILE = "conf/environment.xml";
  private static final String DEFAULT_ENV_CONFIG_RESOURCE = "environment.xml";

  private List<ServiceRegistration> environmentRegistrations;

  protected Factory engineFactory;

  private List<ComponentInstance> instances;

  public OrchestraWSEngineActivator(final BundleContext context, final Class< ? >... extensions) {
    super(context, extensions);
  }

  @Override
  public synchronized void start() throws Exception {
    super.start();
    final Map<String, URL> environmentUrls = new HashMap<String, URL>();
    final File userEnvFile = new File(OrchestraWSEngineActivator.DEFAULT_USER_CONFIG_FILE);
    if (userEnvFile.exists()) {
      final Properties userProperties = new Properties();
      final FileInputStream fip = new FileInputStream(userEnvFile);
      try {
        userProperties.load(fip);
      } finally {
        fip.close();
      }
      for (Object user : userProperties.keySet()) {
        File userEnv = new File(userEnvFile.getParentFile(), userProperties.getProperty((String) user));
        if (userEnv.exists()) {
          try {
            environmentUrls.put((String) user, userEnv.toURI().toURL());
          } catch (final MalformedURLException e) {
            Misc.unreachableStatement();
          }
        }
      }
    } else {
      URL envUrl = null;
      final File envFile = new File(OrchestraWSEngineActivator.DEFAULT_ENV_CONFIG_FILE);
      if (envFile.exists()) {
        try {
          envUrl = envFile.toURI().toURL();
        } catch (final MalformedURLException e) {
          Misc.unreachableStatement();
        }
      }
      if (envUrl == null) {
        envUrl = this.context.getBundle().getResource(OrchestraWSEngineActivator.DEFAULT_ENV_CONFIG_RESOURCE);
      }
      environmentUrls.put("default", envUrl);
    }
    this.environmentRegistrations = new ArrayList<ServiceRegistration>();
    this.instances = new ArrayList<ComponentInstance>();
    for (Entry<String, URL> envUrl : environmentUrls.entrySet()) {
      Misc.fastDynamicLog(OrchestraWSEngineActivator.LOG, Level.INFO,
          "Orchestra instance will be started for user %s using environment from: %s", envUrl.getKey(), envUrl.getValue());

      UUID instanceUUID = UUID.randomUUID();
      Properties properties = new Properties();
      properties.put("environment-id", envUrl.getKey() + "-" + instanceUUID.toString());
      this.environmentRegistrations.add(this.context.registerService(
          EnvironmentProviderService.class.getName(),
          new EnvironmentProvider(envUrl.getValue(), this.context), properties));
      Properties instanceProperties = new Properties();
      instanceProperties.put("requires.filters",
          new String[] {
              "environmentProvider", "(environment-id=" + properties.getProperty("environment-id") + ")"
          });

      this.instances.add(getEngineFactory().createComponentInstance(instanceProperties));
    }
  }

  @Override
  public synchronized void stop() throws Exception {
    if (this.instances != null) {
      for (ComponentInstance instance : instances) {
        instance.dispose();
      }
      instances = null;
    }
    if (this.environmentRegistrations != null) {
      for (ServiceRegistration environmentRegistration : environmentRegistrations) {
        environmentRegistration.unregister();
      }
      this.environmentRegistrations = null;
    }
    super.stop();
  }

  public Factory getEngineFactory() {
    return engineFactory;
  }
  public void setEngineFactory(Factory engineFactory) {
    this.engineFactory = engineFactory;
  }
}
