package org.ow2.orchestra.parsing;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.FaultHandlerThatExits;
import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.extension.ExtensionProvider;
import org.ow2.orchestra.facade.def.ExtensionDefinition;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.ProcessFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.ExtensionDefinitionImpl;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.wsdl.WsdlDefinitions;
import org.ow2.orchestra.lang.evaluator.EvaluatorFactory;
import org.ow2.orchestra.parsing.binding.AssignBinding;
import org.ow2.orchestra.parsing.binding.CompensateBinding;
import org.ow2.orchestra.parsing.binding.CompensateScopeBinding;
import org.ow2.orchestra.parsing.binding.DefinitionKeeper;
import org.ow2.orchestra.parsing.binding.EmptyBinding;
import org.ow2.orchestra.parsing.binding.ExitBinding;
import org.ow2.orchestra.parsing.binding.ExtensionActivityBinding;
import org.ow2.orchestra.parsing.binding.FlowBinding;
import org.ow2.orchestra.parsing.binding.ForEachBinding;
import org.ow2.orchestra.parsing.binding.IfBinding;
import org.ow2.orchestra.parsing.binding.InvokeBinding;
import org.ow2.orchestra.parsing.binding.PickBinding;
import org.ow2.orchestra.parsing.binding.ProcessBinding;
import org.ow2.orchestra.parsing.binding.ReceiveBinding;
import org.ow2.orchestra.parsing.binding.RepeatUntilBinding;
import org.ow2.orchestra.parsing.binding.ReplyBinding;
import org.ow2.orchestra.parsing.binding.RethrowBinding;
import org.ow2.orchestra.parsing.binding.ScopeBinding;
import org.ow2.orchestra.parsing.binding.SequenceBinding;
import org.ow2.orchestra.parsing.binding.ThrowBinding;
import org.ow2.orchestra.parsing.binding.ValidateBinding;
import org.ow2.orchestra.parsing.binding.WaitBinding;
import org.ow2.orchestra.parsing.binding.WhileBinding;
import org.ow2.orchestra.parsing.binding.extension.Extension;
import org.ow2.orchestra.parsing.binding.extension.ExtensionsInfo;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.ProcessResourcesRepository;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.UrlEntity;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.wsdl.BpelWsdlParser;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.ow2.orchestra.xml.Bindings;
import org.ow2.orchestra.xml.Entity;
import org.ow2.orchestra.xml.Parse;
import org.ow2.orchestra.xml.Parser;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

import javax.wsdl.Definition;
import javax.wsdl.Operation;
import javax.wsdl.OperationType;
import javax.wsdl.PortType;
import javax.xml.namespace.QName;
import java.io.InputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * parses the dom tree of the bpel document and stores the builds the
 * BpelProcess. The Deployer will make the BpelProcess available in the object
 * stack of the parse.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class BpelParser extends Parser {

  /**
   * path to the directory containing all xsd files.
   */
  private static final String RESSOURCES_DIR = "";
  /**
   * all schema resources used to parse bpel files.
   */
  private static final String[] SCHEMA_RESOURCES = {
          BpelParser.RESSOURCES_DIR + "wsdl.xsd",
          BpelParser.RESSOURCES_DIR + "xml-schema.xsd",
          BpelParser.RESSOURCES_DIR + "ws-bpel_abstract_common_base.xsd",
          BpelParser.RESSOURCES_DIR + "ws-bpel_executable.xsd"
  };

  /**
   * Constants used for JAXP 1.2
   */
  static final String JAXP_SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
  static final String W3C_XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";
  static final String JAXP_SCHEMA_SOURCE = "http://java.sun.com/xml/jaxp/properties/schemaSource";

  public static final String CLASSPATH_URL = "classpath://";

  public static final String CATEGORY_ACTIVITY = "activities";

  /**
   * all bindings used for parsing.
   */
  private static Bindings defaultBindings = BpelParser.getDefaultBindings();

  /**
   * all allowed tag names for process top level activities.
   */
  // public static final Set<String> processActivityTagNames =
  // getProcessActivityTagNames();

  // the default entities are initialized at the bottom of this file.
  private static Map<String, Entity> defaultEntities = BpelParser.getDefaultEntities();

  private BpelWsdlParser wsdlParser = null;

  /**
   * Default constructor.
   */
  public BpelParser() {
    super(BpelParser.defaultBindings, BpelParser.defaultEntities);
    this.wsdlParser = new BpelWsdlParser();
  }

  /**
   * Parses a the dom document representing the bpel file.
   *
   * @param processElement root element of the document : the process element
   * @param parse          parse
   * @return the resulting bpelProcess
   */
  @Override
  public Object parseDocumentElement(final Element processElement, final Parse parse) {
    final BpelProcess bpelProcess = this.parseProcessAttributes(processElement, parse);
    parse.pushObject(bpelProcess);
    try {
      bpelProcess.setResourcesRepository(parse.findObject(ProcessResourcesRepository.class));
      bpelProcess.setWsdlInfos(parse.findObject(WsdlsInfos.class));
      final FaultHandlerThatExits faultHandlerThatExits = new FaultHandlerThatExits();
      faultHandlerThatExits.setExceptionClassName(BpelFaultException.class.getName());
      bpelProcess.setExceptionHandler(faultHandlerThatExits);

      final List<Element> elements = XmlUtil.elements(processElement);
      if (elements == null || elements.isEmpty()) {
        parse.addProblem("no elements in process", processElement);
        return bpelProcess;
      }

      this.parseProcessElements(processElement, bpelProcess, parse);

      // creates a scope associated to the process : in the bindings : process
      // is mapped to ScopeBinding
      // let's create the top level node for the bpel process.

      final Node processScope = (Node) this.parseElement(processElement, parse, BpelParser.CATEGORY_ACTIVITY);
      bpelProcess.setInitial(processScope);

      // Copy process resources in processFullDefinition
      final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
      definitionKeeper.getProcessFullDefinition().setProcessResources(bpelProcess.getResourcesRepository().getResources());

    } finally {
      parse.popObject();
    }
    return bpelProcess;
  }

  /**
   * Parses all process specific attributes (those not common with scope).
   *
   * @param processElement process element (dom)
   * @param parse          parse
   */
  private BpelProcess parseProcessAttributes(final Element processElement, final Parse parse) {
    final String name = XmlUtil.attribute(processElement, "name");

    final String targetNamespace = XmlUtil.attribute(processElement, "targetNamespace");

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);

    final long version = definitionKeeper.getProcessId();
    final BpelProcess bpelProcess = new BpelProcess(name, targetNamespace, version);

    final String queryLanguage = XmlUtil.attribute(processElement, "queryLanguage");
    if (!EvaluatorFactory.isSupportedLanguage(queryLanguage)) {
      // SA00004
      final String message = StaticAnalysisFault.getMessage("SA00004",
              "  Query language '" + queryLanguage + "' not supported.");
      parse.addProblem(message, new StaticAnalysisException("SA00004"), processElement);
      throw new OrchestraRuntimeException(message);
    }
    bpelProcess.setQueryLanguage(queryLanguage);

    final String expressionLanguage = XmlUtil.attribute(processElement, "expressionLanguage");
    if (!EvaluatorFactory.isSupportedLanguage(expressionLanguage)) {
      // SA00004
      final String message = StaticAnalysisFault.getMessage("SA00004",
              "  Expression language '" + expressionLanguage + "' not supported.");
      parse.addProblem(message, new StaticAnalysisException("SA00004"), processElement);
      throw new OrchestraRuntimeException(message);
    }
    bpelProcess.setExpressionLanguage(expressionLanguage);

    final ProcessDefinitionUUID processDefinitionUUID = bpelProcess.getUUID();

    final ProcessFullDefinition processFullDefinition = new ProcessFullDefinitionImpl(processDefinitionUUID,
            targetNamespace, name, queryLanguage,
            expressionLanguage, new Date());

    definitionKeeper.setProcessFullDefinition(processFullDefinition);

    return bpelProcess;
  }

  /**
   * Parses all process specific children elements (those not common with
   * scope).
   *
   * @param processElement list of all process children elements
   * @param bpelProcess    bpelProcess
   * @param parse          parse
   */
  private void parseProcessElements(final Element processElement, final BpelProcess bpelProcess, final Parse parse) {
    final Element extensions = XmlUtil.element(processElement, "extensions");
    if (extensions != null) {
      List<Element> listExtension = XmlUtil.elements(extensions, "extension");
      if (listExtension != null) {
        for (final Element extension : listExtension) {
          this.parseExtensionElement(extension, bpelProcess, parse);
        }
        //add list of extensions definition to process full Definition
        final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
        ExtensionsInfo extInfo = parse.findObject(ExtensionsInfo.class);
        for (Extension extDef : extInfo.getListDefinition()) {
          ExtensionDefinition extensionDefinition = new ExtensionDefinitionImpl(extDef.getNamespace(), extDef.isMustUnderstand(), extDef.isSupported());
          definitionKeeper.getProcessFullDefinition().getExtensions().add(extensionDefinition);
        }

      } else {
        parse.addProblem("missing extension definition in extensions " + XmlUtil.toString(extensions), extensions);
      }
    }
    final List<Element> importElements = XmlUtil.elements(processElement, "import");
    if (importElements != null) {
      for (final Element importElement : importElements) {
        this.parseImportElement(importElement, bpelProcess, parse);
      }
    }
  }

  /**
   * Parse Bpel process extension element.
   */
  private void parseExtensionElement(final Element extension, final BpelProcess bpelProcess, final Parse parse) {

    String namespace = null;
    boolean mustUnderstand = false;
    if (extension.hasAttribute("namespace")) {
      namespace = extension.getAttribute("namespace");
    }

    if (!extension.hasAttribute("mustUnderstand")) {
      parse.addProblem("missing mustUnderstand attribute in extension element " + XmlUtil.toString(extension), extension);
    } else {
      mustUnderstand = extension.getAttribute("mustUnderstand").equals("yes");
    }

    if (namespace == null) {
      parse.addProblem("missing namespace attribute in extension element " + XmlUtil.toString(extension), extension);
    } else {
      // get the list of extension provider and this extension is supported if we have an extension provider which support it
      final List<ExtensionProvider> extensionProviders = EnvTool.get(List.class, "extensionProviders");
      boolean isSupported = false;
      if (extensionProviders != null) {
        for (ExtensionProvider extensionProvider : extensionProviders) {
          if (extensionProvider.isSupported(namespace)) {
            isSupported = true;
          }
        }
      }
      parse.findObject(ExtensionsInfo.class).addDefinition(new Extension(namespace, mustUnderstand, isSupported));
    }

  }

  /**
   * Parse Bpel process import element.
   */
  private void parseImportElement(final Element importElement, final BpelProcess bpelProcess, final Parse parse) {
    String namespace = null;
    if (importElement.hasAttribute("namespace")) {
      namespace = importElement.getAttribute("namespace");
    }
    /*
    * String location = null; if (importElement.hasAttribute("location")) {
    * location = importElement.getAttribute("location"); }
    */

    String importType = null;
    if (importElement.hasAttribute("importType")) {
      importType = importElement.getAttribute("importType");
    }

    if (importType == null) {
      parse.addProblem("missing importType attribute in import element " + XmlUtil.toString(importElement), importElement);
    } else if (importType.equals(XmlConstants.XMLNS_WSDL)) {
      // Import a WSDL from Deployment
      final WsdlDefinitions wsdlDefinitions = parse.findObject(WsdlDefinitions.class);
      Set<Definition> definitionsForNS = null;
      if (wsdlDefinitions != null) {
        definitionsForNS = wsdlDefinitions.getWsdlDefinitions(namespace);
      }
      if (definitionsForNS == null) {
        parse.addProblem("No wsdl file to import in import element " + XmlUtil.toString(importElement), importElement);
      } else {
        for (final Definition wsdlDef : definitionsForNS) {
          // perform static analysis
          if (namespace == null && wsdlDef.getTargetNamespace() != null) {
            parse.addProblem("invalid namespace in imported wsdl (SA00012) :"
                    + " no target namespace expected, got \"" + wsdlDef.getTargetNamespace() + "\"", importElement);
          } else if (namespace != null && !namespace.equals(wsdlDef.getTargetNamespace())) {
            parse
                    .addProblem("invalid namespace in imported wsdl (SA00011) :" + " expected \""
                            + namespace + "\", got \"" + wsdlDef.getTargetNamespace() + "\"", importElement);
          }

          // Static Analysis
          final Map<QName, PortType> portTypes = wsdlDef.getAllPortTypes();
          for (final PortType pt : portTypes.values()) {
            final List<Operation> operations = pt.getOperations();
            for (int index = 0; index < operations.size(); index++) {
              final Operation op = operations.get(index);
              final OperationType opType = op.getStyle();
              // SA00001
              if (opType == null || opType.equals(OperationType.NOTIFICATION)
                      || opType.equals(OperationType.SOLICIT_RESPONSE)) {
                final String message;
                if (opType == null) {
                  message = StaticAnalysisFault.getMessage("SA00001",
                          "  PortType: '" + pt.getQName().toString() + "', Operation: '" + op.getName()
                                  + "' has unsupported operation type: (unknown type)");
                } else {
                  message = StaticAnalysisFault.getMessage("SA00001",
                          "  PortType: '" + pt.getQName().toString() + "', Operation: '" + op.getName()
                                  + "' has unsupported operation type: " + opType.toString());
                }
                parse.addProblem(message, new StaticAnalysisException("SA00001"), importElement);
              }
              final String name = op.getName();
              int cmpt = 1;
              for (int i = index + 1; i < operations.size(); i++) {
                if (operations.get(i).getName().equals(name)) {
                  cmpt++;
                }
              }
              if (cmpt > 1) {
                // SA00002
                final String message = StaticAnalysisFault.getMessage("SA00002",
                        "  PartnerLink " + pt.getQName().toString() + " has " + cmpt
                                + " operations with name " + name);
                parse.addProblem(message, new StaticAnalysisException("SA00002"), importElement);
              }
            }
          }

          this.wsdlParser.parse(wsdlDef, parse, bpelProcess.getQueryLanguage(), bpelProcess, bpelProcess.getWsdlInfos());
        }
      }
    } else {
      parse.addProblem("unsupported importType attribute in import element " + XmlUtil.toString(importElement),
              importElement.getAttributeNode("importType"));
    }
  }

  /**
   * Returns all necessary bindings that the parser should use.
   *
   * @return the build Bindings object
   */
  public static Bindings getDefaultBindings() {
    final Bindings bindings = new Bindings();
    bindings.addBinding(new ProcessBinding());
    bindings.addBinding(new SequenceBinding());
    bindings.addBinding(new ScopeBinding());
    bindings.addBinding(new ReceiveBinding());
    bindings.addBinding(new ReplyBinding());
    bindings.addBinding(new FlowBinding());
    bindings.addBinding(new EmptyBinding());
    bindings.addBinding(new AssignBinding());
    bindings.addBinding(new InvokeBinding());
    bindings.addBinding(new IfBinding());
    bindings.addBinding(new CompensateBinding());
    bindings.addBinding(new CompensateScopeBinding());
    bindings.addBinding(new ExitBinding());
    bindings.addBinding(new ForEachBinding());
    bindings.addBinding(new PickBinding());
    bindings.addBinding(new RepeatUntilBinding());
    bindings.addBinding(new ThrowBinding());
    bindings.addBinding(new RethrowBinding());
    bindings.addBinding(new ValidateBinding());
    bindings.addBinding(new WaitBinding());
    bindings.addBinding(new WhileBinding());

    bindings.addBinding(new ExtensionActivityBinding());
    return bindings;
  }

  protected static QName getBPELQName(final String localName) {
    return new QName(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, localName);
  }

  public static Set<QName> getAllActivityQNames() {
    final Set<String> allActivityTagNames = BpelParser.getDefaultBindings().getTagNames(BpelParser.CATEGORY_ACTIVITY);
    final Set<QName> allActivityQNames = new HashSet<QName>();
    for (final String tagName : allActivityTagNames) {
      allActivityQNames.add(BpelParser.getBPELQName(tagName));
    }
    return allActivityQNames;
  }

  /**
   * Returns all necessary entities that the parser should use.
   *
   * @return a map containing all entities : key : url on the internet, value :
   *         local resource.
   */
  private static Map<String, Entity> getDefaultEntities() {
    final Map<String, Entity> defaultSchemaCatalog = new HashMap<String, Entity>();
    final ClassLoader resourceLoader = BpelParser.class.getClassLoader();
    defaultSchemaCatalog.put("http://schemas.xmlsoap.org/wsdl/",
            new UrlEntity(BpelParser.RESSOURCES_DIR + "wsdl.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://www.w3.org/2001/XMLSchema",
            new UrlEntity(BpelParser.RESSOURCES_DIR + "xml-schema.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://www.w3.org/2001/xml.xsd",
            new UrlEntity(BpelParser.RESSOURCES_DIR + "xml-schema.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://docs.oasis-open.org/wsbpel/2.0/process/abstract",
            new UrlEntity(BpelParser.RESSOURCES_DIR + "ws-bpel_abstract_common_base.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://docs.oasis-open.org/wsbpel/2.0/process/executable",
            new UrlEntity(BpelParser.RESSOURCES_DIR + "ws-bpel_executable.xsd",
                    resourceLoader));
    return defaultSchemaCatalog;
  }

  /**
   * Build a string array containing all schema resources to use when parsing a
   * bpel file.
   *
   * @return the build string array
   */
  @Override
  protected String[] getSchemaSources() {
    final String[] schemas = new String[BpelParser.SCHEMA_RESOURCES.length];
    for (int i = 0; i < BpelParser.SCHEMA_RESOURCES.length; i++) {
      schemas[i] = BpelParser.CLASSPATH_URL + BpelParser.SCHEMA_RESOURCES[i];
    }
    return schemas;
  }

  /*
  * (non-Javadoc)
  *
  * @see org.ow2.orchestra.xml.Parser#resolveEntity(java.lang.String,
  * java.lang.String)
  */
  @Override
  public InputSource resolveEntity(final String publicId, final String systemId) {
    if (systemId.startsWith(BpelParser.CLASSPATH_URL)) {
      final String localpart = systemId.substring(BpelParser.CLASSPATH_URL.length());
      final InputStream inStream = Thread.currentThread().getContextClassLoader().getResourceAsStream(localpart);
      if (inStream != null) {
        return new InputSource(inStream);
      }
    }
    return super.resolveEntity(publicId, systemId);
  }

}
