/**
 * Copyright (C) 2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.definition.activity.Compensate;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.ActivityDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.CompensateActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.xml.Parse;
import org.ow2.orchestra.xml.Parser;
import org.w3c.dom.Element;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author S. Ali Tokmen
 */
public class CompensateBinding extends ActivityBinding {

  /**
   * logger.
   */
  private static Logger log = Logger.getLogger(CompensateBinding.class.getName());

  public CompensateBinding() {
    super("compensate", ActivityType.COMPENSATE);
  }

  /**
   * Parses the given dom element to build a Node and the Compensate behavior
   *
   * @param element dom element to parse
   * @param parse   parse
   * @param parser  parser
   * @return the Compensate behavior
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (CompensateBinding.log.isLoggable(Level.FINE)) {
      CompensateBinding.log.fine("parsing element = " + element);
    }

    final Compensate compensate = new Compensate();
    this.setEnclosingScope(compensate, parse);
    this.parseStandardAttributes(element, compensate, parse);
    ActivityBinding.parseExtensionAttributes(element, compensate, parse);
    this.parseStandardElements(element, compensate, parse);
    ActivityBinding.parseExtensionElements(element, compensate, parse);

    final Node activityNode = this.createActivityNode(parse, parser, compensate);

    // Static Analysis SA00007
    boolean foundFHorCHorTH = false;
    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    for (final Object parent : definitionKeeper.getParentsStack()) {
      final ActivityDefinition parentActivity = (ActivityDefinition) parent;
      final ActivityType parentActivityType = parentActivity.getType();
      if (parentActivityType == ActivityType.CATCH_HANDLER || parentActivityType == ActivityType.CATCHALL_HANDLER
              || parentActivityType == ActivityType.COMPENSATION_HANDLER || parentActivityType == ActivityType.TERMINATION_HANDLER) {
        foundFHorCHorTH = true;
        break;
      }
    }
    if (!foundFHorCHorTH) {
      final String message = StaticAnalysisFault.getMessage("SA00007", "  The <compensate> element \"" + compensate.getName()
              + "\" is not in a <catch>, <catchAll>, <compensationHandler> or <terminationHandler> element.");
      parse.addProblem(message, new StaticAnalysisException("SA00007"), element);
      return null;
    }

    return activityNode;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
                                                                 final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new CompensateActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
