/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.definition.activity.While;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.WhileActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.WhileActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.lang.evaluator.EvaluatorFactory;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.xml.Parse;
import org.ow2.orchestra.xml.Parser;
import org.w3c.dom.Element;

import java.util.logging.Level;
import java.util.logging.Logger;

public class WhileBinding extends ActivityBinding {

  /**
   * log.
   */
  private static Logger log = Logger.getLogger(IfBinding.class.getName());


  public WhileBinding() {
    super("while", ActivityType.WHILE);
  }

  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (WhileBinding.log.isLoggable(Level.FINE)) {
      WhileBinding.log.fine("parsing element = " + element);
    }
    final While whileObject = new While();
    this.setEnclosingScope(whileObject, parse);
    this.parseStandardAttributes(element, whileObject, parse);
    ActivityBinding.parseExtensionAttributes(element, whileObject, parse);
    this.parseStandardElements(element, whileObject, parse);
    ActivityBinding.parseExtensionElements(element, whileObject, parse);

    final Node whileNode = this.createActivityNode(parse, parser, whileObject);

    parse.pushObject(whileNode);
    parse.pushObject(whileObject);
    try {
      whileObject.setCondition(this.parseCondition(element, parse));
      this.parseChildActivity(element, parse, parser);
    } finally {
      parse.popObject();
      parse.popObject();
    }

    return whileNode;
  }

  private Expression parseCondition(final Element parentElement, final Parse parse) {
    final Element conditionElement = XmlUtil.element(parentElement, "condition");
    if (conditionElement != null) {
      final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
      String expressionLanguage = XmlUtil.attribute(conditionElement,
              "expressionLanguage");
      if (expressionLanguage == null) {
        expressionLanguage = bpelProcess.getExpressionLanguage();
      }
      final Expression expression = new Expression();
      if (!EvaluatorFactory.isSupportedLanguage(expressionLanguage)) {
        // SA00004
        final String message = StaticAnalysisFault.getMessage("SA00004",
                "  Expression language '" + expressionLanguage + "' not supported.");
        parse.addProblem(message, new StaticAnalysisException("SA00004"), parentElement);
      }
      expression.setLanguage(expressionLanguage);
      expression.setNamespaces(BpelUtil.getAllNameSpaces(conditionElement));
      final String textContent = conditionElement.getTextContent();
      expression.setText(textContent);
      final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
      final WhileActivityFullDefinition activityFullDefinition = (WhileActivityFullDefinition) definitionKeeper
              .getActivityFullDefinition();
      activityFullDefinition.setConditionBooleanExpression(textContent);
      activityFullDefinition.setConditionExpressionLanguage(expressionLanguage);
      return expression;
    }
    return null;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
                                                                 final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new WhileActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
