/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.CompositeElement;
import org.ow2.orchestra.definition.ExceptionHandler;
import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.activity.BaseCatchActivity;
import org.ow2.orchestra.definition.activity.CatchActivity;
import org.ow2.orchestra.definition.activity.Compensate;
import org.ow2.orchestra.definition.activity.CompensationHandlerActivity;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.persistence.db.TimerSession;
import org.ow2.orchestra.runtime.op.AtomicOperation;
import org.ow2.orchestra.runtime.op.ExecuteNode;
import org.ow2.orchestra.runtime.op.MoveToChildNode;
import org.ow2.orchestra.runtime.op.MoveToParentNode;
import org.ow2.orchestra.runtime.op.Signal;
import org.ow2.orchestra.services.OpenIMAKey;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.job.Job;
import org.ow2.orchestra.services.job.JobImpl;
import org.ow2.orchestra.services.job.TimerImpl;
import org.ow2.orchestra.services.jobexecutor.JobDbSession;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.Priority;

/**
 * a runtime path of execution.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class BpelExecution {

  public static final String ADDITIONNAL_SCOPE_VARIABLES = "additionnalScopeVariables";
  public static final String SCOPE_STATE_VARIABLE = "scopeState";

  // atomic operations
  public static final AtomicOperation EXECUTE_NODE = new ExecuteNode();
  public static final AtomicOperation PROPAGATE_TO_PARENT = new MoveToParentNode();

  /** log. */
  private static final Logger LOG = Logger.getLogger(BpelExecution.class.getName());

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  protected long dbid;
  protected int dbversion;

  /** a unique id for this execution. */
  protected String id;

  /** @see Execution */
  protected ExecutionState state;


  // current position /////////////////////////////////////////////////////////

  /** current node is not persisted, only its name*/
  protected Node currentNode;
  protected String currentNodeName;

  /** are concurrent executions that related to this execution. */
  protected Collection<BpelExecution> executions;

  /**
   * the parent child relation of executions is convenient for some forms of
   * concurrency.
   */
  protected BpelExecution parent = null;
  protected BpelInstance processInstance;

  protected Map<String, Object> variables = new HashMap<String, Object>();

  protected List<JobImpl< ? >> jobs = new ArrayList<JobImpl< ? >>();

  // transient members

  /** the queue of atomic operations to be performed for this execution. */
  protected Queue<AtomicOperation> atomicOperations;

  protected boolean waitForSignal;

  protected Node previousNode;
  protected Exception exception;

  protected ActivityInstanceUUID activityInstanceUUID;


  protected FlowRuntime flowRuntime;

  protected ForEachRuntime forEachRuntime;

  protected ScopeRuntime scopeRuntime;

  protected WaitingExecution waitingExecution;

  /** State of the execution before the suspend. Used to restore the same state when resuming execution. */
  protected ExecutionState resumeState;

  // Constructor is empty because hibernate will call this constructor
  // each time the object is fetched from the database
  public BpelExecution() { }


  /**
   * Search for the flowRuntime having a link declared with the given name. We
   * start searching into parent and not into this execution as links used in an
   * activity (the current node) must be declared before (in an enclosing flow).
   * So if the current node is flow, we don't want to search into its links. If
   * this parent execution flowRuntime doesn't have one, look into parent
   *
   * @param linkName
   *          linkName
   * @return the flowRuntime having a link declared with the given name.
   */
  public FlowRuntime getFlowRuntimeWithLink(final String linkName) {
    final FlowRuntime fr = this.getFlowRuntime();
    if (fr != null && fr.hasLink(linkName)) {
      return fr;
    }
    if (this.getParent() != null) {
      return (this.getParent()).getFlowRuntimeWithLink(linkName);
    }
    return null;
  }

  public ScopeRuntime getCurrentScopeRuntime() {
    if (this.scopeRuntime != null) {
      return this.scopeRuntime;
    }
    if (this.parent != null) {
      return (this.getParent()).getCurrentScopeRuntime();
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithVariable(final String variableName) {
    if (this.scopeRuntime != null) {
      if (this.scopeRuntime.containsVariable(variableName)) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithVariable(variableName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithCS(final String csName) {
    if (this.scopeRuntime != null) {
      final Scope scope = (Scope) this.getProcessDefinition().getNode(this.scopeRuntime.getScopeNodeUUID()).getBehaviour();
      if (scope.getCorrelationSets().containsKey(csName)) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithCS(csName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithPLNK(final String partnerLinkName) {
    if (this.scopeRuntime != null) {
      final Scope scope = (Scope) this.getProcessDefinition().getNode(this.scopeRuntime.getScopeNodeUUID()).getBehaviour();
      if (scope.getPartnerLinks().containsKey(partnerLinkName)) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithPLNK(partnerLinkName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithScope(final Scope scope) {
    if (this.scopeRuntime != null) {
      if (this.scopeRuntime.getScopeNodeUUID().equals(scope.getUUID().toString())) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithScope(scope);
    }
    return null;
  }

  /**
   * toString method.
   *
   * @return String string representation of this execution.
   */
  @Override
  public String toString() {
    final StringBuffer sb = new StringBuffer("BpelExecution");

    if (this.getNode() != null) {
      final AbstractActivity activity = this.getNode().getBehaviour();
      sb.append('[');
      sb.append(activity.getName());
      sb.append(']');
    }

    sb.append(this.hashCode());
    return sb.toString();
  }

  /**
   * The process of this execution is guaranteed to be a {@link BpelProcess}.
   */
  public BpelProcess getProcessDefinition() {
    return this.getProcessInstance().getProcessDefinition();
  }

  /**
   * Search in currentScope (and parents) for a partnerLinkRuntime having the
   * given name.
   *
   * @param partnerLinkName
   *          partnerLinkName of the searched partnerLinkRuntime
   * @return the partnerLinkRuntime having the given partnerLinkName in the
   *         closest enclosing scope.
   */
  public PartnerLinkRuntime getPartnerLinkRuntime(final String partnerLinkName) {
    final ScopeRuntime scopeRuntimeWithPLNK = this.getScopeRuntimeWithPLNK(partnerLinkName);
    if (scopeRuntimeWithPLNK != null) {
      return scopeRuntimeWithPLNK.getPartnerLink(partnerLinkName);
    }
    return null;
  }

  /**
   *
   * /** Search for a variableRuntime having the given variableName in the
   * closest scope and return its value.
   *
   * @param variableName
   *          variableName
   * @return Search for a variableRuntime having the given variableName in the
   *         closest scope and return its value.
   */
  public Object getVariableValue(final String variableName) {
    final VariableRuntime variable = this.getVariableByName(variableName);
    return variable.getValue();
  }

  /**
   * Search for a variableRuntime having the given variableName in the closest
   * scope.
   *
   * @param variableName
   *          variableName
   * @return the found variableRuntime
   */
  public VariableRuntime getVariableByName(final String variableName) {
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINE, "Getting variable %s", variableName);
    final ScopeRuntime currentRuntimeScope = this.getScopeRuntimeWithVariable(variableName);
    if (currentRuntimeScope != null) {
      return currentRuntimeScope.getVariable(variableName);
    }
    return null;
  }

  /**
   * @param variableName
   * @return
   */
  public Variable getVariableDefinition(final String variableName) {
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINE, "Getting variable definition %s", variableName);
    final AbstractActivity activity = this.getNode().getBehaviour();
    switch (activity.getType()) {
      case SCOPE:
        final Variable var = ((Scope) activity).getVariable(variableName);
        if (var != null) {
          return var;
        }
        break;
      case CATCH_HANDLER:
        final Variable var2 = ((CatchActivity) activity).getVariable(variableName);
        if (var2 != null) {
          return var2;
        }
        break;
      default:
    }
    if (this.scopeRuntime != null && this.scopeRuntime.containsVariable(variableName)) {
      final Scope scope = (Scope) this.getProcessDefinition().getNode(this.scopeRuntime.getScopeNodeUUID()).getBehaviour();
      return scope.getVariable(variableName);
    }
    if (this.getParent() != null) {
      return this.getParent().getVariableDefinition(variableName);
    }
    return null;
  }
  /**
   * Search for a correlationSetRuntime having the given csName in the closest
   * scope.
   *
   * @param csName
   *          csName
   * @return the found correlationSetRuntime
   */
  public CorrelationSetRuntime getCorrelationSetByName(final String csName) {
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINE, "Getting CorrelationSet %s", csName);
    final ScopeRuntime scopeRuntimeWithCS = this.getScopeRuntimeWithCS(csName);
    if (scopeRuntimeWithCS != null) {
      return scopeRuntimeWithCS.getCorrelationSet(csName);
    }
    return null;
  }

  public void pushFlowRuntime(final FlowRuntime flowRuntimeToPush) {
    if (this.flowRuntime == null) {
      this.flowRuntime = flowRuntimeToPush;
    } else {
      Misc.unreachableStatement("One execution should have only one flowRuntime runtime");
    }
  }

  public FlowRuntime getFlowRuntime() {
    return this.flowRuntime;
  }

  /**
   * @param forEachRuntimeToPush
   *          the forEachRuntime to push
   */
  public void pushForEachRuntime(final ForEachRuntime forEachRuntimeToPush) {
    if (this.forEachRuntime != null) {
      Misc.unreachableStatement("One execution should have only one foreach runtime");
    }
    this.forEachRuntime = forEachRuntimeToPush;
  }

  /**
   * Get the forEachRuntime.
   */
  public ForEachRuntime getForEachRuntime() {
    return this.forEachRuntime;
  }

  /**
   * @param scopeRuntimeToPush
   *          the scopeRuntime to push
   */
  public void pushScopeRuntime(final ScopeRuntime scopeRuntimeToPush) {
    if (this.scopeRuntime != null) {
      Misc.unreachableStatement("One execution should have only one scopeRuntime");
    }
    this.scopeRuntime = scopeRuntimeToPush;
    EnvTool.getRepository().storeScopeRuntime(scopeRuntimeToPush);
  }

  /**
   * Pop a scopeRuntime.
   */
  public ScopeRuntime popScopeRuntime(final boolean shouldAdd) {
    if (this.scopeRuntime == null) {
      Misc.unreachableStatement("There is no scopeRuntime to pop!");
    }
    final ScopeRuntime res = this.scopeRuntime;
    this.scopeRuntime = null;
    final ScopeRuntime parentScopeRuntime = this.getCurrentScopeRuntime();
    final Node chNode = this.getNode().getNodes().get(Scope.COMPENSATION_ACTIVITY);
    final CompensationHandlerActivity chActivity = (CompensationHandlerActivity) chNode.getBehaviour();
    // Add scope to list of ended children
    // only if there is a compensation handler defined or if there are child scopes to compensate.
    if (shouldAdd && (!chActivity.isDefaultCH() || !res.getEndedChildren().isEmpty())) {
      if (parentScopeRuntime == null) {
        Misc.unreachableStatement("parent is null");
        return null;
      }
      parentScopeRuntime.addEndedChild(res);
    } else {
      // don't keep the scope runtime
      if (parentScopeRuntime != null) {
        parentScopeRuntime.getEndedChildren().remove(res);
      }
      EnvTool.getRepository().removeScopeRuntime(res);
    }
    return res;
  }

  /**
   * Get the scopeRuntime.
   */
  public ScopeRuntime getScopeRuntime() {
    return this.scopeRuntime;
  }

  public Replier popReplier(final OperationKey operationKey, final String messageExchange) {
    if (this.scopeRuntime != null) {
      final MessageExchangeRuntime messageExchangeRuntime = new MessageExchangeRuntime(messageExchange, this.scopeRuntime);
      final OpenIMAKey openIMAKey = new OpenIMAKey(operationKey, messageExchangeRuntime);
      final Iterator<Replier> replierIterator =  this.scopeRuntime.getRepliers().iterator();
      while (replierIterator.hasNext()) {
        final Replier replier = replierIterator.next();
        if (replier.getOpenIMAKey().equals(openIMAKey)) {
          replierIterator.remove();
          return replier;
        }
      }
    }
    if (this.parent != null) {
      return this.getParent().popReplier(operationKey, messageExchange);
    }
    return null;
  }

  public void setWaitingExecution(final WaitingExecution waitingExecution) {
    this.waitingExecution = waitingExecution;
    EnvTool.getRepository().storeWaitingExecution(waitingExecution);
  }

  public void resetWaitingExecution() {
    if (this.getWaitingExecution() != null) {
      EnvTool.getRepository().removeWaitingExecution(this.getWaitingExecution());
    }
    this.waitingExecution = null;
  }

  public WaitingExecution getWaitingExecution() {
    return this.waitingExecution;
  }

  public boolean hasWaitingExecution() {
    return this.waitingExecution != null
      && this.waitingExecution.getInboundMessageElementName() != null
      && !"".equals(this.waitingExecution.getInboundMessageElementName());
  }

  @Override
  public int hashCode() {
    return this.id.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof BpelExecution)) {
      return obj.equals(this);
    }
    final BpelExecution other = (BpelExecution) obj;
    return other.getId().equals(this.id);
  }


  public Set<BpelExecution> searchWaitingExecutions() {
    final Set<BpelExecution> res = new HashSet<BpelExecution>();
    if (this.hasWaitingExecution()) {
      res.add(this);
    }
    if (this.getExecutions() != null) {
      for (final BpelExecution child : this.getExecutions()) {
        res.addAll(child.searchWaitingExecutions());
      }
    }
    return res;
  }

  /**
   * Method called to cleanly terminate child activities. If a child activity is
   * a scope, then invokes the termination handler.
   */
  public void terminate() {
    final ActivityType activityType = this.getNode().getBehaviour().getType();
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST,
        "Terminating BpelExecution %s, activity type %s, node name %s", this, activityType, this.getNode().getName());


    if (this.scopeRuntime != null) {
      if (!ActivityType.SCOPE.equals(activityType)) {
        Misc.unreachableStatement("The execution " + this + " has a scopeRuntime but is not a "
            + ActivityType.SCOPE + " (it is a " + activityType + ")");
      } else {
        Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST,
            "Scope state for BpelExecution %s is %s", this, this.scopeRuntime.getState());
      }
    } else if (ActivityType.SCOPE.equals(activityType)) {
      Misc.unreachableStatement("The execution " + this + " doen't have any scopeRuntime but is a " + activityType);
    }

    final Collection<BpelExecution> childExecutions = this.getExecutions();
    if (childExecutions != null && !childExecutions.isEmpty()) {
      // If there's a child, terminate just one child and then return.
      // the child will call terminate on the parent when child termination is completed.
      final BpelExecution childBpelExecution = childExecutions.iterator().next();
      Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST,
          "Terminating child %s out of %s children", childBpelExecution, childExecutions.size());
      childBpelExecution.terminate();
      return;
    } else if (ActivityType.SCOPE.equals(activityType)
        && ScopeState.HANDLING_FAULT.equals(this.scopeRuntime.getState())) {
      // All enclosed activities have been terminated,
      // execute catch
      this.executeFaultHandler();
      return;
    } else if (ActivityType.SCOPE.equals(activityType)
        && (ScopeState.RUNNING.equals(this.scopeRuntime.getState())
            /*|| ScopeState.RETHROWING_FAULT.equals(this.scopeRuntime.getState())*/)) {
      // if there is no more child in a scope, execute its termination handler
      // then return.
      // the termination handler will call terminate again if we are in a nested
      // scope.
      this.executeTerminationHandler();
      return;
    } else if (ActivityType.FOR_EACH.equals(activityType) && this.getForEachRuntime() != null
      && this.getForEachRuntime().isCompleted()) {
      // for each:
      // if forEach is completed (not terminated because of a fault),
      // all remaining branches have been terminated
      // continue normal execution.
      this.signal("all-finished");
      return;
    } else if (ActivityType.COMPENSATION_HANDLER.equals(activityType) && this.getParent().getScopeRuntime() != null) {
      // compensation handler terminated
      this.signalCompensationHandler();
    } else if (ActivityType.SCOPE_EVENT_HANDLER.equals(activityType)) {
      // if there is no more child in another activity, just exit the activity
      // and propagate to parent.
      // Note that the propagation is not called when:
      // - in the scope handling a fault
      // - in the forEach activity that is successfully completed
      // This seems to be the only activities which causes termination to
      // called.
      this.terminateEventHandler();
    } else {
      // if there is no more child in another activity, just exit the activity
      // and propagate to parent.
      // Note that the propagation is not called when:
      // - in the scope handling a fault
      // - in the forEach activity that is successfully completed
      // This seems to be the only activities which causes termination to
      // called.
      this.terminateCurrentActivity();
    }
  }


  private void terminateEventHandler() {
    // DO NOT RECORD termination, because this is only a virtual activity (not monitored).
    this.removeExecution();
    if (this.getState().equals(ExecutionState.async)) {
      this.unlock();
      this.end(ExecutionState.ended);
    } else {
      this.end(ExecutionState.cancelled);
    }
    if (this.parent != null) {
      Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "Terminating parent %s", this.parent);
      (this.parent).terminate();
    }
  }


  private void terminateCurrentActivity() {
    this.removeExecution();
    if (this.getState().equals(ExecutionState.async) || this.getState().equals(ExecutionState.waitingLinks)) {
      this.unlock();
      this.end(ExecutionState.ended);
    } else {
      this.end(ExecutionState.cancelled);
      if (this.parent != null) {
        if (this.parent.getNode() != this.getNode()) {
          final AbstractActivity activity = this.getNode().getBehaviour();
          EnvTool.getRecorder().recordActivityTerminated(activity.createRuntimeData(this));
        }
      }
    }
    if (this.parent != null) {
      Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "Terminating parent %s", this.parent);
      (this.parent).terminate();
    }
  }


  private void signalCompensationHandler() {
    if (this.getParent().getScopeRuntime().getLastException() != null) {
      final BpelExecution chExecution = this.getParent();
      // remove finished execution from parent
      chExecution.removeExecution(this);
      // end ch execution
      chExecution.end(ExecutionState.ended);

      // remove scopeRuntime
      final ScopeRuntime chScopeRuntime = chExecution.popScopeRuntime(false);
      chScopeRuntime.setState(ScopeState.COMPENSATED, this);
      final BpelExecution compensateExecution = chScopeRuntime.getCompensateExecution();
      chScopeRuntime.setCompensateExecution(null);

      // remove finished execution from parent
      final BpelExecution parentExecution = chExecution.getParent();
      parentExecution.removeExecution(chExecution);

      // Signal calling compensate activity
      final Map<String,Object> params = new HashMap<String, Object>();
      params.put(Compensate.EXCEPTION_PARAMETER, chScopeRuntime.getLastException());
      compensateExecution.signal(params);
    }
  }


  private void executeTerminationHandler() {
    this.scopeRuntime.setState(ScopeState.TERMINATING, this);

    // Get to the Scope NodeImpl and all children
    final List<Node> nodes = this.getNode().getNodes();
    if (nodes == null) {
      Misc.unreachableStatement("The scope node " + this.getNode() + " doesn't have any children");
      return;
    }
    if (nodes.size() < Scope.BIGGEST_ACTIVITY_INDEX) {
      Misc.unreachableStatement("The scope node " + this.getNode() + " has only " + nodes.size()
        + " children, it should have had at least " + Scope.BIGGEST_ACTIVITY_INDEX);
    }

    final Node terminationHandler = nodes.get(Scope.TERMINATION_ACTIVITY);
    final BpelExecution terminationHandlerExecution = this.createExecution();

    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST,
        "Will execute the termination handler %s in BpelExecution %s for the BpelExecution %s",
        terminationHandler, terminationHandlerExecution, this);
    terminationHandlerExecution.execute(terminationHandler);
  }


  private void executeFaultHandler() {
    // Get to the Scope NodeImpl and all children
    final List<Node> nodes = this.getNode().getNodes();
    if (nodes == null) {
      Misc.unreachableStatement("The parent scope node doesn't have any children");
      return;
    }
    if (nodes.size() < Scope.BIGGEST_ACTIVITY_INDEX) {
      Misc.unreachableStatement("The parent scope node has only " + nodes.size() + " children, it should have had at least "
        + Scope.BIGGEST_ACTIVITY_INDEX);
    }

    Node faultHandler = BaseCatchActivity.findFaultHandler(this.scopeRuntime, nodes, this.scopeRuntime.getLastException());
    if (faultHandler == null) {
      // If no appropriate fault handler matched, use the catchAll handler
      // (wsbpel-v2.0-OS, page 129, parts 1 and 2, last 2 cases for both)
      faultHandler = nodes.get(Scope.CATCHALL_ACTIVITY);
    }

    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST,
        "Will execute the fault handler %s in BpelExecution %s", faultHandler, this);
    this.createExecution().execute(faultHandler);
  }

  /**
   * Method called to abruptly stop all child activities. If a child activity is
   * a scope, then the termination handler is NOT invoked.
   */
  public void exit() {
    if (this.isSuspended()) {
      this.setState(this.resumeState);
    }
    // Remove execution first to avoid getting called twice
    this.removeExecution();

    // Exit all children
    final Collection<BpelExecution> childExecutions = this.getExecutions();
    if (childExecutions != null) {
      for (final BpelExecution childExecution : new ArrayList<BpelExecution>(childExecutions)) {
        childExecution.exit();
      }
    }
    if (this.getState().equals(ExecutionState.async) || this.getState().equals(ExecutionState.waitingLinks)) {
      this.unlock();
      this.end(ExecutionState.ended);
    } else if (!this.isFinished()) {
      this.end(ExecutionState.cancelled);
      if (this.getParent() != null && this.getParent().getNode() != this.getNode()) {
        final AbstractActivity activity = this.getNode().getBehaviour();
        EnvTool.getRecorder().recordActivityExited(activity.createRuntimeData(this));
      }
    }
  }

  protected void removeExecution() {
    // Delete execution from all receivers
    this.resetWaitingExecution();
    // remove jobs
    this.removeJobs();

    if (this.parent != null) {
      // Make sure this execution doesn't call back its parent
      this.waitForSignal();
      this.parent.removeExecution(this);
    }
  }

  /**
   * Get the current activity instance uuid.
   * @return the activityInstanceUUID
   */
  public ActivityInstanceUUID getActivityInstanceUUID() {
    return this.activityInstanceUUID;
  }

  /**
   * Get the activity instance uuid of the encloser activity of the current activity.
   * @return the activityInstanceUUID
   */
  public ActivityInstanceUUID getEncloserActivityInstanceUUID() {
    return this.getEncloserActivityInstanceUUID(this.getActivityInstanceUUID());
  }

  /**
   * @return the activityInstanceUUID
   */
  private ActivityInstanceUUID getEncloserActivityInstanceUUID(final ActivityInstanceUUID enclosedActivityUUID) {
    final BpelExecution parentBpelExecution = this.getParent();
    // use getter instead of direct field access.
    // direct field access creates a bug with Hibernate
    if (enclosedActivityUUID.equals(this.getActivityInstanceUUID())) {
      if (parentBpelExecution == null) {
        return null;
      } else {
        return parentBpelExecution.getEncloserActivityInstanceUUID(enclosedActivityUUID);
      }
    } else {
      return this.getActivityInstanceUUID();
    }
  }

  /**
   * @param activityInstanceUUID
   *          the activityInstanceUUID to set
   */
  public void setActivityInstanceUUID(final ActivityInstanceUUID activityInstanceUUID) {
    this.activityInstanceUUID = activityInstanceUUID;
  }




  public void removeJobs() {
    final JobDbSession jobSession = Environment.getFromCurrent(JobDbSession.class);
    for (final Job job : new HashSet<Job>(this.getJobs())) {
      jobSession.delete(job);
    }
  }


  /**
   * @return the resumeState
   */
  public ExecutionState getResumeState() {
    return this.resumeState;
  }


  /**
   * @param resumeState the resumeState to set
   */
  public void setResumeState(final ExecutionState resumeState) {
    this.resumeState = resumeState;
  }

  // From executionImpl




  // execution method : end ///////////////////////////////////////////////////

  public void end() {
    this.end(ExecutionState.ended);
  }

  public void end(final ExecutionState endState) {
    if (endState == null) {
      throw new OrchestraRuntimeException("state is null");
    }
    if (endState.equals(ExecutionState.active)
        || endState.equals(ExecutionState.created)
        || endState.equals(ExecutionState.inactive)
        || endState.equals(ExecutionState.suspended)
        || endState.equals(ExecutionState.async)) {
      throw new OrchestraRuntimeException("invalid end state: " + endState);
    }

    if (endState.equals(ExecutionState.ended)) {
      Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "%s ends", this);
    } else {
      Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "%s ends with state %s", this, endState);
    }

    // end all child executions
    if (this.executions != null) {
      for (final BpelExecution child : this.executions) {
        child.end(endState);
      }
    }

    this.lock(endState);

    this.waitForSignal = true;
  }

  // execution method : signal ////////////////////////////////////////////////

  public void signal() {
    this.signal(null, (Map<String, Object>) null);
  }

  public void signal(final String signal) {
    this.signal(signal, (Map<String, Object>) null);
  }

  public void signal(final Map<String, Object> parameters) {
    this.signal(null, parameters);
  }

  public void signal(final String signal, final Map<String, Object> parameters) {
    this.checkLock();
    if (this.getNode() != null) {
      this.performAtomicOperation(new Signal(signal, parameters, this.getNode()));
    } else {
      throw new OrchestraRuntimeException("execution is not in a node");
    }
  }

  // execution method : execute ///////////////////////////////////////////////

  /** @see Execution#execute(Node) */
  public void execute(final Node nodeToExecute) {
    if (nodeToExecute == null) {
      throw new OrchestraRuntimeException("node is null");
    }
    this.checkLock();

    this.waitForSignal = true;
    this.performAtomicOperation(new MoveToChildNode(nodeToExecute));
  }

  // execution method : waitForSignal /////////////////////////////////////////

  public void waitForSignal() {
    this.waitForSignal = true;
  }

  // execution method : proceed ///////////////////////////////////////////////

  public void proceed() {
    this.checkLock();

    // in block structured processDefinition languages we assume that
    // there is a parent node of the current node
    final Node parentNode = this.getNode().getParentNode();

    // if there is a parent node
    if (parentNode != null) {
      // propagate to the parent
      this.performAtomicOperation(BpelExecution.PROPAGATE_TO_PARENT);

    } else {
      // When we don't know how to proceed, i don't know if it's best to
      // throw new PvmException("don't know how to proceed");
      // or to end the execution. Because of convenience for testing,
      // I opted to end the execution.
      this.end();
    }
  }

  public void move(final Node destination, final BpelExecution execution) {
    execution.move(destination);
  }

  public void move(final Node destination) {
    this.checkLock();
    this.setNode(destination);
  }

  // execution : internal methods /////////////////////////////////////////////

  public void moveTo(final Node destination) {
    // if the parent node needs to know the previous node
    this.setPreviousNode(this.getNode());

    // move the execution to the destination
    this.setNode(destination);
  }


  // variables ////////////////////////////////////////////////////////////////

  public void createVariable(final String variableKey, final Object variableValue) {
    if (this.isFinished()) {
      throw new OrchestraRuntimeException("can't create variable '" + variableKey + "' on " + this
          + ": " + this.state);
    }

    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST,
        "create variable '%s' in '%s' with value '%s'", variableKey, this, variableValue);

    this.variables.put(variableKey, variableValue);

  }

  public Object getVariable(final String variableKey) {
    return this.variables.get(variableKey);
  }



  public boolean removeVariable(final String variableKey) {
    if (this.isFinished()) {
      throw new OrchestraRuntimeException("can't remove variable '" + variableKey + "' on " + this
          + ": " + this.state);
    }
    return this.variables.remove(variableKey) != null;
  }

  public void createTimer(final String signalName, final Date dueDate, final Long repeat) {
    if (signalName == null) {
      throw new OrchestraRuntimeException("no signal specified");
    }
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "creating timer on %s", this);

    // instantiate the timer
    final TimerImpl timerImpl = this.instantiateTimer();
    // set priority
    timerImpl.setPriority(Priority.HIGHEST);
    // create the bidirectional reference
    timerImpl.setExecution(this);
    this.jobs.add(timerImpl);
    // setInverseReference(timerImpl);

    // initialise all the timer properties
    timerImpl.setSignalName(signalName);
    if (dueDate != null) {
      timerImpl.setDueDate(dueDate);
    }

    // the if is added to keep the original default
    if (repeat != null) {
      timerImpl.setRepeat(repeat);
    }

    // get the TimerSession from the environment
    final Environment environment = Environment.getCurrent();
    if (environment == null) {
      throw new OrchestraRuntimeException("non environment for initializing timers");
    }
    final TimerSession timerSession = environment.get(TimerSession.class);
    if (timerSession == null) {
      throw new OrchestraRuntimeException(
          "no TimerSession in environment for initializing timers");
    }

    // schedule the timer with the TimerSession
    timerSession.schedule(timerImpl);
  }

  public List<JobImpl<?>> getJobs() {
    return this.jobs;
  }

  protected TimerImpl instantiateTimer() {
    return new TimerImpl();
  }

  // state ////////////////////////////////////////////////////////////////////

  /** @see Execution#getState() */
  public ExecutionState getState() {
    return this.state;
  }

  /** @see Execution#lock(String) */
  public void lock(final ExecutionState lockState) {
    if (lockState == null) {
      throw new OrchestraRuntimeException("given state is null");
    }
    this.checkLock();
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "locking %s", this);
    this.state = lockState;
  }

  public void unlock() {
    if (ExecutionState.active.equals(this.state)) {
      throw new OrchestraRuntimeException("state is already active");
    }
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "unlocking %s", this);
    this.state = ExecutionState.active;
  }

  public boolean isActive() {
    return ExecutionState.active.equals(this.state);
  }

  public boolean isLocked() {
    return !this.isActive();
  }

  public boolean isSuspended() {
    return ExecutionState.suspended.equals(this.state);
  }

  public boolean isEnded() {
    return ExecutionState.ended.equals(this.state);
  }

  public boolean isFinished() {
    return ExecutionState.ended.equals(this.state) || ExecutionState.cancelled.equals(this.state);
  }

  // state : internal methods /////////////////////////////////////////////////

  protected void checkLock() {
    if (!ExecutionState.active.equals(this.state)) {
      throw new OrchestraRuntimeException(this.toString() + " is not active: " + this.state);
    }
  }

  //  continuations ////////////////////////////////////////////////


  public void performAtomicOperation(final AtomicOperation operation) {
    if (this.atomicOperations == null) {

      // initialise the fifo queue of atomic operations
      this.atomicOperations = new LinkedList<AtomicOperation>();
      this.atomicOperations.add(operation);

      try {
        while (!this.atomicOperations.isEmpty()) {
          final AtomicOperation atomicOperation = this.atomicOperations.remove();
          atomicOperation.perform(this);
        }

      } catch (final RuntimeException e) {
        throw e;
      } finally {
        this.atomicOperations = null;
      }
    } else {
      this.atomicOperations.add(operation);
    }
  }

  // exception handling

  public void handleException(CompositeElement observableElement,
      Exception thrownException, final String rethrowMessage) {

    final List<CompositeElement> processElements = new ArrayList<CompositeElement>();
    while (observableElement != null) {
      processElements.add(observableElement);
      observableElement = observableElement.getParent();
    }

    for (final CompositeElement processElement : processElements) {
      final ExceptionHandler exceptionHandler = processElement.getExceptionHandler();
      if (exceptionHandler != null) {
        if (exceptionHandler.matches(thrownException)) {
          try {
            exceptionHandler.handle(this, thrownException, processElement);
            return;
          } catch (final Exception rethrowException) {
            if (!exceptionHandler.isRethrowMasked()) {
              thrownException = rethrowException;
            }
          }
        }
      }
    }

    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST,
        "rethrowing exception cause no exception handler for %s", thrownException);
    ExceptionHandler.rethrow(thrownException, rethrowMessage + ": "
        + thrownException.getMessage());
  }

  // child executions /////////////////////////////////////////////////////////

  /** @see Execution#createProcessInstance() */
  public BpelExecution createExecution() {
    // creating a child execution implies that this execution
    // is not a leave any more and therefore, it is inactivated
    if (this.isActive()) {
      this.lock(ExecutionState.inactive);
      this.waitForSignal = true;
    }

    // create child execution
    final BpelExecution childExecution = new BpelExecution();
    childExecution.id = UUID.randomUUID().toString();
    childExecution.processInstance = this.processInstance;
    childExecution.setNode(this.getNode());
    childExecution.state = ExecutionState.active;
    Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "creating %s", childExecution);
    // add it to this execution
    this.addExecution(childExecution);

    childExecution.setActivityInstanceUUID(this.activityInstanceUUID);
    EnvTool.getRepository().storeExecution(childExecution);

    return childExecution;
  }

  public void addExecution(final BpelExecution execution) {
    final BpelExecution executionImpl = execution;
    executionImpl.parent = this;
    if (this.executions == null) {
      this.executions = new ArrayList<BpelExecution>();
    }
    this.executions.add(executionImpl);
  }

  public void removeExecution(final BpelExecution child) {
    if (this.executions != null) {
      if (this.executions.remove(child)) {
        if (this.state.equals(ExecutionState.inactive) && (this.executions.isEmpty())) {
          Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "last child execution was removed; unlocking");
          this.state = ExecutionState.active;
        } else {
          Misc.fastDynamicLog(BpelExecution.LOG, Level.FINEST, "removed %s from %s", child, this);
        }
      } else {
        throw new OrchestraRuntimeException(child + " is not a child execution of " + this);
      }
    }
    // Delete execution from all receivers
    child.resetWaitingExecution();
    // remove jobs
    child.removeJobs();
    EnvTool.getRepository().removeExecution(child);
  }

  // node name ////////////////////////////////////////////////////////////////

  public String getNodeName() {
    return this.currentNodeName;
  }

  // getters and setters
  // /////////////////////////////////////////////////////////



  @SuppressWarnings("unchecked")
  public Collection<BpelExecution> getExecutions() {
    return this.executions;
  }

  public BpelExecution getParent() {
    return this.parent;
  }

  public Node getPreviousNode() {
    return this.previousNode;
  }

  /**
   * Returns the rootExecution (instance).
   *
   * @return the rootExecution (instance)
   */
  public BpelInstance getProcessInstance() {
    return this.processInstance;
  }

  public void setProcessInstance(final BpelInstance processInstance) {
    this.processInstance = processInstance;
  }


  public Exception getException() {
    return this.exception;
  }

  public void setException(final Exception exception) {
    this.exception = exception;
  }

  public boolean isWaitForSignal() {
    return this.waitForSignal;
  }

  public void setWaitForSignal(final boolean waitForSignal) {
    this.waitForSignal = waitForSignal;
  }

  public void setState(final ExecutionState state) {
    this.state = state;
  }

  public void setExecutions(final Collection<BpelExecution> executions) {
    this.executions = executions;
  }

  public void setParent(final BpelExecution parent) {
    this.parent = parent;
  }

  public void setPreviousNode(final Node previousNode) {
    this.previousNode = previousNode;
  }

  public Node getNode() {
    if (this.currentNode == null) {
      final BpelProcess processDef = this.getProcessDefinition();
      this.currentNode = processDef.getNode(this.getCurrentNodeName());
      if (this.currentNode == null) {
        System.out.println("No node found for " + this.currentNodeName + ". Known nodes:" + processDef.getNodesMap().keySet());
      }
    }
    return this.currentNode;
  }

  public void setNode(final Node node) {
    this.currentNode = node;
    this.currentNodeName = node.getName();
  }

  public long getDbid() {
    return this.dbid;
  }

  public String getId() {
    return this.id;
  }

  public void setId(final String id) {
    this.id = id;
  }

  /**
   * @return the currentNodeName
   */
  public String getCurrentNodeName() {
    return this.currentNodeName;
  }

}
