/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import javax.xml.namespace.QName;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.EqualsUtil;
import org.ow2.orchestra.util.QNameComparator;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.var.MessageVariable;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * @author Charles Souillard
 */
public class PendingMessage {

  public static interface PendingMessageIterator {
    PendingMessage next();
  }

  protected long dbid;
  protected int dbversion;

  protected MessageVariable message;

  protected OperationKey operationKey;

  protected String messageCarrierId;

  protected Date addedDate;

  // List of correlation keys (one for each receiving element that can handle the message)
  protected List<Integer> correlationKeys = new ArrayList<Integer>();

  protected PendingMessage() {
    super();
  }

  protected PendingMessage(final MessageVariable message, final String messageCarrierId, final OperationKey operationKey) {
    super();
    this.message = message;
    this.messageCarrierId = messageCarrierId;
    this.operationKey = operationKey;
    this.addedDate = new Date();
  }

  protected void fillCorrelationKeys(List<BpelProcess> processes) {
    List<ReceivingElement> receivingElements = new ArrayList<ReceivingElement>();
    for (BpelProcess process : processes) {
      for (final ReceivingElement receivingElement : process.getReceivingElements(this.operationKey)) {
        final QName messageType = receivingElement.getInputMessageType();
        final SortedMap<QName, Document> csProperties = getCSPropertiesFromMessage(process, receivingElement, messageType);
        final List<Set<QName>> propertiesNamesPermutations = ReceivingElementUtil.getCorrelationPropertyNamesPermutations(receivingElement);

        for (final Set<QName> propQNames : propertiesNamesPermutations) {
          final Map<QName, Document> csProps = new HashMap<QName, Document>(csProperties);
          csProps.keySet().retainAll(propQNames);
          final int correlationKey = ReceivingElementUtil.getCorrelationPropertiesKey(csProps);
          this.correlationKeys.add(correlationKey);
        }
      }
    }
  }

  private SortedMap<QName, Document> getCSPropertiesFromMessage(final BpelProcess process, final ReceivingElement receivingElement, final QName messageType) {
    final SortedMap<QName, Document> csProperties = new TreeMap<QName, Document>(QNameComparator.INSTANCE);
    for (final Correlation correlation : receivingElement.getCorrelations()) {
      if (correlation.getInitiate().equals(Initiate.no) || correlation.getInitiate().equals(Initiate.join)) {
        final List<PropertyAlias> propertyAliases =
                process.getWsdlInfos()
                        .getPropertyAliases(correlation.getCorrelationSet().getProperties().values(), messageType);
        //foreach propertyAlias, we extract the value from the message.
        for (final PropertyAlias propertyAlias : propertyAliases) {
          try {
            final Element partElement = message.getPartValue(propertyAlias.getPart());

            Node value = partElement;
            if (propertyAlias.getQuery() != null) {
              value = propertyAlias.getQuery().getEvaluator().evaluate(partElement);
            }
            final Element el = XmlUtil.getDocumentWithOneElement(
                    new QName(XmlConstants.ELT_DEFAULT_NS, XmlConstants.ELT_DEFAULT_ELEMENT));
            el.appendChild(el.getOwnerDocument().importNode(value, true));
            csProperties.put(propertyAlias.getPropertyName(), el.getOwnerDocument());
          } catch (final Exception ex) {
            throw new OrchestraRuntimeException("In Correlation: Function call exception:" + ex, ex);
          }
        }
      }
    }
    return csProperties;
  }

  @Override
  public String toString() {
    return "message = " + this.message.toString() + ", messageCarrier = "
        + this.messageCarrierId;
  }

  public MessageVariable getMessage() {
    return this.message;
  }

  public String getMessageCarrierId() {
    return this.messageCarrierId;
  }

  public OperationKey getOperationKey() {
    return this.operationKey;
  }

  /**
   * @return the addedDate
   */
  public Date getAddedDate() {
    return this.addedDate;
  }

  public long getDbid() {
    return this.dbid;
  }

  // equals ///////////////////////////////////////////////////////////////////
  // hack to support comparing hibernate proxies against the real objects
  // since this always falls back to ==, we don't need to overwrite the hashcode
  @Override
  public boolean equals(final Object o) {
    return EqualsUtil.equals(this, o);
  }
  @Override
  public int hashCode() {
    return super.hashCode();
  }
}
