/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import org.ow2.orchestra.facade.Record;
import org.ow2.orchestra.facade.criteria.Criteria;
import org.ow2.orchestra.facade.criteria.ProcessInstancesSortableFields;
import org.ow2.orchestra.facade.def.ActivityDefinition;
import org.ow2.orchestra.facade.def.ProcessState;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.runtime.ActivityInstance;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ProcessInstance;
import org.ow2.orchestra.facade.runtime.full.ActivityFullInstance;
import org.ow2.orchestra.facade.runtime.full.ProcessFullInstance;
import org.ow2.orchestra.facade.uuid.AbstractUUID;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.services.Querier;
import org.ow2.orchestra.services.QuerierPagedResult;
import org.ow2.orchestra.util.Misc;

import javax.xml.namespace.QName;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Guillaume Porcher
 *         <p/>
 *         Chainer for Queriers : - for methods that returns only one object,
 *         search for the first matching object - for methods that returns a
 *         collection: search for all objects in all return queriers.
 */
public class QuerierChainer implements Querier {

  private static final Logger LOG = Logger.getLogger(QuerierChainer.class.getName());

  private final List<Querier> queriers;

  private <E extends Record> void putRecordCollection(final Map<AbstractUUID, E> map, final Collection<E> recordCollection) {
    if (recordCollection != null) {
      for (final E record : recordCollection) {
        AbstractUUID uuid = null;
        if (record instanceof ActivityInstance) {
          uuid = ((ActivityInstance) record).getUUID();
        } else if (record instanceof ActivityDefinition) {
          uuid = ((ActivityDefinition) record).getUUID();
        }
        if (map.put(uuid, record) != null) {
          Misc.fastDynamicLog(QuerierChainer.LOG, Level.WARNING,
                  "Duplicated entry (same uuid:%s) found. Keep only the newest", uuid);
        }
      }
    }
  }

  private static class DeploymentDateComparator implements Comparator<ProcessFullDefinition>, Serializable {

    private static final long serialVersionUID = 2463131104245876514L;

    public int compare(final ProcessFullDefinition o1, final ProcessFullDefinition o2) {
      final long o1date = o1.getDeployedDate() != null ? o1.getDeployedDate().getTime() : 0;
      final long o2date = o2.getDeployedDate() != null ? o2.getDeployedDate().getTime() : 0;
      if (o1date == o2date) {
        return 0;
      } else if (o1date < o2date) {
        return -1;
      } else {
        return 1;
      }
    }

  }

  public QuerierChainer(final List<Querier> queriers) {
    this.queriers = queriers;
  }

  public Set<ActivityFullInstance> findActivityInstances(final ProcessInstanceUUID processUUID) {
    final Map<AbstractUUID, ActivityFullInstance> activityFullInstanceByUUIDMap =
            new HashMap<AbstractUUID, ActivityFullInstance>();
    final Set<ActivityFullInstance> activityInsts = new HashSet<ActivityFullInstance>();
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(activityFullInstanceByUUIDMap, querier.findActivityInstances(processUUID));
    }
    activityInsts.addAll(activityFullInstanceByUUIDMap.values());
    return activityInsts;
  }

  public Set<ActivityFullInstance> findActivityInstances(final ProcessInstanceUUID processUUID,
                                                         final ActivityState activityState) {
    final Map<AbstractUUID, ActivityFullInstance> activityFullInstanceByUUIDMap =
            new HashMap<AbstractUUID, ActivityFullInstance>();
    final Set<ActivityFullInstance> activityInsts = new HashSet<ActivityFullInstance>();
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(activityFullInstanceByUUIDMap, querier.findActivityInstances(processUUID, activityState));
    }
    activityInsts.addAll(activityFullInstanceByUUIDMap.values());
    return activityInsts;
  }

  public Set<ProcessFullInstance> findProcessInstances() {
    final Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessInstances());
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  /**
   * Retrieve the UUID of a record.
   *
   * @param record the record
   * @param <E>    One of {@link ActivityInstance} or {@link ActivityDefinition}
   * @return UUID of the given record
   */
  private <E extends Record> AbstractUUID getUuidFromRecord(final E record) {
    AbstractUUID uuid = null;
    if (record instanceof ActivityInstance) {
      uuid = ((ActivityInstance) record).getUUID();
    } else if (record instanceof ActivityDefinition) {
      uuid = ((ActivityDefinition) record).getUUID();
    }
    return uuid;
  }

  /**
   * <p>Merges records into a map.<br>
   * If a is contained in both map and recordCollection then the one within the recordCollection will overwrite the one within
   * the map.<br>
   * A {@link LinkedHashMap} is used to preserve insertion order.
   * </p>
   *
   * @param map              the map potentially containing existing records into which records will be merged
   * @param recordCollection Collection of new records to merge
   * @param modulo           modulo used to compute insertion index
   * @param <E>              One of {@link ActivityInstance} or {@link ActivityDefinition}
   * @return Returns the number of duplicate records found during merge
   */
  private <E extends Record> int mergePages(final LinkedHashMap<AbstractUUID, E> map, final Collection<E> recordCollection,
                                             final int modulo) {

    int nbDuplicate = 0;
    if (recordCollection.size() <= 0) {
      return nbDuplicate;
    }

    LinkedHashMap<AbstractUUID, E> copy = new LinkedHashMap<AbstractUUID, E>(map);
    map.clear();
    //First remove duplicates if any
    for (final E record : recordCollection) {
      AbstractUUID uuid = getUuidFromRecord(record);
      if (copy.containsKey(uuid)) {
        copy.remove(uuid);
        nbDuplicate++;
      }
    }
    //then merge both lists
    Iterator<E> records = recordCollection.iterator();
    Iterator<E> existingElements = copy.values().iterator();
    for (int i = 1; i <= copy.values().size(); i++) {
      final E existingRecord = existingElements.next();
      map.put(getUuidFromRecord(existingRecord), existingRecord);

      if (i % modulo == 0 && records.hasNext()) {
        final E record = records.next();
        map.put(getUuidFromRecord(record), record);
      }
    }

    //add remaining records to merge
    while (records.hasNext()) {
      final E record = records.next();
      map.put(getUuidFromRecord(record), record);
    }
   return nbDuplicate;
  }

  public QuerierPagedResult<ProcessFullInstance> findProcessInstances(final Criteria<ProcessInstance> criteria) {
    final QuerierPagedResult<ProcessFullInstance> result = new QuerierPagedResult<ProcessFullInstance>();
    //Linked hash map ensure that keys are iterated in the order of their insertion.
    final LinkedHashMap<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new LinkedHashMap<AbstractUUID,
            ProcessFullInstance>();
    List<ProcessFullInstance> resultList = new ArrayList<ProcessFullInstance>();

    // Paging over multiples database is not so easy.
    // If there is only one querier then there is no problem and we use pagination directly.
    // Otherwise we retrieve a page starting at index 0 to firstIndex + maxPageSize on each database.
    // then those results are merged together and potentially reordered according criteria.
    // finally a new page is retrieved within this merged-reordered collection.
    // This causes last pages retrieval to be less efficient than for first pages.

    Criteria<ProcessInstance> newCriteria;
    if (this.queriers.size() > 1) {
      newCriteria = new Criteria<ProcessInstance>().setOrder(criteria.getOrder()).setFirstResult(0);
      if (criteria.getMaxResults() > 0) {
        newCriteria.setMaxResults(criteria.getFirstResult() + criteria.getMaxResults());
      }
      newCriteria.getRestrictions().addAll(criteria.getRestrictions());
    } else {
      newCriteria = criteria;
    }
    // Workaround end

    int modulo = 1;
    for (final Querier querier : this.queriers) {
      final QuerierPagedResult<ProcessFullInstance> page = querier.findProcessInstances(newCriteria);
      if (page != null) {
        result.setTotalElementsCount(result.getTotalElementsCount()
                + page.getTotalElementsCount()
                - this.mergePages(processFullInstanceByUUIDMap, page.getElements(), modulo));
        modulo++;
      }
    }
    resultList.addAll(processFullInstanceByUUIDMap.values());

    // Workaround
    if (this.queriers.size() > 1) {

      //Reorder collections
      if (criteria.getOrder() != null) {
        Collections.sort(resultList, new Comparator<ProcessFullInstance>() {

          public int compare(ProcessFullInstance o1, ProcessFullInstance o2) {
            if (o1 == o2) {
              return 0;
            }

            int diff = -1;
            if (criteria.getOrder().getField().equals(ProcessInstancesSortableFields.LAST_UPDATE_DATE)) {
              if (o1.getProcessInstanceStateUpdate().size() == 0 && o2.getProcessInstanceStateUpdate().size() != 0) {
                diff = -1;
              } else if (o2.getProcessInstanceStateUpdate().size() == 0 && o1.getProcessInstanceStateUpdate().size() != 0) {
                diff = 1;
              } else if (o1.getProcessInstanceStateUpdate().size() == 0 && o2.getProcessInstanceStateUpdate().size() == 0) {
                diff = 0;
              } else {
                diff = o1
                        .getProcessInstanceStateUpdate().get(o1.getProcessInstanceStateUpdate().size() - 1).getDate()
                        .compareTo(o2.getProcessInstanceStateUpdate().get(o2.getProcessInstanceStateUpdate().size() - 1).getDate());
              }
            } else if (criteria.getOrder().getField().equals(ProcessInstancesSortableFields.START_DATE)) {
              if (o1.getStartedDate() != null) {
                diff = (o2.getStartedDate() != null) ? o1.getStartedDate().compareTo(o2.getStartedDate()) : 1;
              }
            } else if (criteria.getOrder().getField().equals(ProcessInstancesSortableFields.STATE)) {
              if (o1.getState() != null) {
                diff = (o2.getState() != null) ? o1.getState().toString().compareTo(o2.getState().toString()) : 1;
              }
            }
            return criteria.getOrder().isAscending() ? diff : -diff;
          }
        });
      }

      //Deal with the page start
      if (criteria.getFirstResult() > 0) {
        if (criteria.getFirstResult() > resultList.size()) {
          resultList.clear();
        } else {
          resultList = resultList.subList(criteria.getFirstResult(), resultList.size());
        }
      }

      //Deal with the page length
      if (criteria.getMaxResults() > 0 && resultList.size() > criteria.getMaxResults()) {
        result.getElements().addAll(resultList.subList(0, criteria.getMaxResults()));
      } else {
        result.getElements().addAll(resultList);
      }

    } else {
      //Simply use result list
      result.getElements().addAll(resultList);
    }
    // End Workaround
    return result;
  }

  public Set<ProcessFullInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID) {
    final Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessInstances(processDefinitionUUID));
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  public Set<ProcessFullInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID,
                                                       final ActivityState processInstanceState) {
    final Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap,
              querier.findProcessInstances(processDefinitionUUID, processInstanceState));
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  public Set<ProcessFullInstance> findProcessInstances(final ActivityState processInstanceState) {
    final Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessInstances(processInstanceState));
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  public ProcessFullInstance getProcessInstance(final ProcessInstanceUUID processInstanceUUID) {
    for (final Querier querier : this.queriers) {
      final ProcessFullInstance processFullInstance =
              querier.getProcessInstance(processInstanceUUID);
      if (processFullInstance != null) {
        return processFullInstance;
      }
    }
    return null;
  }

  public ActivityFullInstance getActivityInstance(final ActivityInstanceUUID activityInstanceUUID) {
    for (final Querier querier : this.queriers) {
      final ActivityFullInstance activityFullInstance = querier.getActivityInstance(activityInstanceUUID);
      if (activityFullInstance != null) {
        return activityFullInstance;
      }
    }
    return null;
  }

  public ActivityFullDefinition getActivityDefinition(final ActivityDefinitionUUID activtyDefinitionUUID) {
    for (final Querier querier : this.queriers) {
      final ActivityFullDefinition activityFullInstance = querier.getActivityDefinition(activtyDefinitionUUID);
      if (activityFullInstance != null) {
        return activityFullInstance;
      }
    }
    return null;
  }

  public Set<ActivityFullDefinition> findActivityDefinitions(final ProcessDefinitionUUID processUUID) {
    final Map<AbstractUUID, ActivityFullDefinition> activityFullDefByUUIDMap =
            new HashMap<AbstractUUID, ActivityFullDefinition>();
    final Set<ActivityFullDefinition> activityDefs = new HashSet<ActivityFullDefinition>();
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(activityFullDefByUUIDMap, querier.findActivityDefinitions(processUUID));
    }
    activityDefs.addAll(activityFullDefByUUIDMap.values());
    return activityDefs;
  }

  public List<ProcessFullDefinition> findProcessDefinitions(final QName name, final ProcessState state) {
    final Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap =
            new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions(name, state));
    }
    final List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public List<ProcessFullDefinition> findProcessDefinitions(final QName name) {
    final Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap =
            new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions(name));
    }
    final List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public List<ProcessFullDefinition> findProcessDefinitions(final ProcessState state) {
    final Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap =
            new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions(state));
    }
    final List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public List<ProcessFullDefinition> findProcessDefinitions() {
    final Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap =
            new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      this.putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions());
    }
    final List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public ProcessFullDefinition getProcessDefinition(final ProcessDefinitionUUID processDefinitionUUID) {
    for (final Querier querier : this.queriers) {
      final ProcessFullDefinition processFullInstance = querier.getProcessDefinition(processDefinitionUUID);
      if (processFullInstance != null) {
        return processFullInstance;
      }
    }
    return null;
  }

  public ProcessFullDefinition removeProcessDefinition(final ProcessDefinitionUUID processDefinitionUUID) {
    for (final Querier querier : this.queriers) {
      final ProcessFullDefinition processFullDefinition = querier.removeProcessDefinition(processDefinitionUUID);
      if (processFullDefinition != null) {
        return processFullDefinition;
      }
    }
    return null;
  }

  public ProcessFullInstance removeProcessInstance(final ProcessInstanceUUID processInstanceUUID) {
    for (final Querier querier : this.queriers) {
      final ProcessFullInstance processFullInstance = querier.removeProcessInstance(processInstanceUUID);
      if (processFullInstance != null) {
        return processFullInstance;
      }
    }
    return null;
  }
}
