/**
 * Copyright (C) 2008 Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/

package org.ow2.orchestra.services.job;

import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.ExecElementToSignal;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.SignalExecUtil;
import org.ow2.orchestra.services.jobexecutor.JobAddedNotification;
import org.ow2.orchestra.services.jobexecutor.JobDbSession;
import org.ow2.orchestra.services.jobexecutor.JobExecutor;
import org.ow2.orchestra.util.Clock;
import org.ow2.orchestra.util.Misc;

/**
 * Try to find a message that can be given to the current execution.
 *
 * @author Guillaume Porcher
 *
 */
public class CheckHasMessageJob extends MessageImpl<Boolean> {

  private static final long serialVersionUID = 3671423433628407972L;
  private static final Logger LOG = Logger.getLogger(CheckHasMessageJob.class.getName());

  private static final int RETRIES = 3;
  private int repeat = CheckHasMessageJob.RETRIES;

  protected CheckHasMessageJob() { }
  public CheckHasMessageJob(final BpelExecution execution) {
    this.setExecution(execution);
  }

  public Boolean execute(final Environment environment) {
    final BpelExecution bpelExecution = this.execution;
    ExecElementToSignal execElt = null;
    // check if execution is still up-to-date
    if (bpelExecution.hasWaitingExecution()) {
      // find a new child execution to signal
      execElt = Receiver.findMessageForExecution(bpelExecution);
    } else {
      Misc.fastDynamicLog(CheckHasMessageJob.LOG, Level.FINEST,
          "execution is not waiting anymore: %s", bpelExecution);
    }
    // repeat one time in case the message arrived during this job execution
    if (this.repeat > 0 && execElt == null) {
      // there is a repeat on this timer
      // then the repeat action dueDate could already have passed
      this.setDueDate(new Date(Clock.getCurrentTime().getTime() + 5000 * (CheckHasMessageJob.RETRIES - this.repeat + 1)));
      Misc.fastDynamicLog(CheckHasMessageJob.LOG, Level.FINEST,
          "rescheduled %s for %s", this, TimerImpl.formatDueDate(this.dueDate));

      // release the lock on the job
      this.setLockOwner(null);
      this.setLockExpirationTime(null);

      // notify the jobExecutor at the end of the transaction
      final JobExecutor jobExecutor = environment.get(JobExecutor.class);
      if (jobExecutor != null) {
        final Transaction transaction = environment.get(Transaction.class);
        if (transaction == null) {
          throw new OrchestraRuntimeException("no transaction in environment");
        }
        final JobAddedNotification jobNotificator = new JobAddedNotification(jobExecutor, this);
        transaction.registerSynchronization(jobNotificator);
      }
      this.repeat = this.repeat - 1;
      return Boolean.FALSE;
    }

    if (execElt != null) {
      // create signal job
      SignalExecUtil.execute(
          environment,
          execElt.getBpelExecution(),
          execElt.getReceivingElement(),
          execElt.getPendingMessage());
    }
    // delete job
    final JobDbSession dbSession = environment.get(JobDbSession.class);
    if (dbSession == null) {
      throw new OrchestraRuntimeException("no " + JobDbSession.class.getName() + " in environment");
    }
    dbSession.delete(this);
    return Boolean.TRUE;

  }

}
