/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.jobexecutor;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.persistence.db.TimerSession;
import org.ow2.orchestra.services.job.Timer;
import org.ow2.orchestra.services.job.TimerImpl;
import org.ow2.orchestra.util.Misc;

/**
 * Timers created with this service are committed at the end of the transaction,
 * so their execution will be late if the delay is shorter than the transaction.
 * In that case, they will be executed at the end of the transaction.
 *
 * @author Tom Baeyens, Pascal Verdage
 */
public class JobExecutorTimerSession implements TimerSession {

  private static final Logger LOG = Logger.getLogger(JobExecutorTimerSession.class.getName());

  /* injected */
  private Transaction transaction;

  /* injected */
  private JobExecutor jobExecutor;

  /* injected */
  private JobDbSession jobDbSession;

  /* injected */
  private int defaultRetries = 3;

  public void schedule(final TimerImpl timer) {
    Misc.fastDynamicLog(JobExecutorTimerSession.LOG, Level.FINE, "scheduling %s", timer);
    this.validate(timer);

    if (timer.isDefaultRetries()) {
      timer.setRetries(this.defaultRetries);
    }

    this.jobDbSession.save(timer);

    this.transaction
          .registerSynchronization(new JobAddedNotification(this.jobExecutor, timer));
  }

  protected void validate(final Timer timer) {
    if (timer == null) {
      throw new OrchestraRuntimeException("null timer scheduled");
    }
    if (timer.getExecution() == null) {
      throw new OrchestraRuntimeException("timer has no execution specified");
    }
    if (timer.getSignalName() == null) {
      throw new OrchestraRuntimeException("timer has no signalName specified");
    }
    if (timer.getDueDate() == null) {
      throw new OrchestraRuntimeException("timer scheduled at null date");
    }
    if (timer.getDueDate().getTime() < 0) {
      throw new OrchestraRuntimeException("timer scheduled with a negative date");
    }
  }

  public void cancel(final TimerImpl timer) {
    Misc.fastDynamicLog(JobExecutorTimerSession.LOG, Level.FINE, "canceling %s", timer);
    if (timer != null) {
      this.jobDbSession.delete(timer);
    } else {
      throw new OrchestraRuntimeException("timer is null");
    }
  }


  /**
   * @param defaultRetries the defaultRetries to set
   */
  public void setDefaultRetries(final int defaultRetries) {
    this.defaultRetries = defaultRetries;
  }
}
